package com.jdwsoftware.wfh.control;


import java.awt.Graphics;
import java.awt.Image;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Cursor;
import java.awt.Color;
import java.awt.Toolkit;
import java.awt.image.PixelGrabber;
import java.awt.image.MemoryImageSource;

import java.util.Random;

import com.jdwsoftware.wfh.model.Fish;
import com.jdwsoftware.wfh.model.Direction;
import com.jdwsoftware.wfh.model.Location;
import com.jdwsoftware.wfh.model.GameType;
import com.jdwsoftware.wfh.model.ColorGame;
import com.jdwsoftware.wfh.model.NumberGame;
import com.jdwsoftware.wfh.model.MathGame;
import javax.swing.*;

/**
 * <p><b>Title</b>: Woody's Fishing Hole</p>
 * <p><b>Description</b>: Educational Fishing Game</p>
 * <p><b>Copyright</b>: Copyright (C) 2003 JDW Software, LLC</p>
 * <p><b>Company</b>: JDW Software, LLC</p>
 *
 * <p>This is the heart of the game.  This component displays most
 *    of the graphics, handles game logic, creates fish, pretty much
 *    all the logic for the game.</p>
 *
 * <p>This class consists of 2 major parts, the main class which
 *    sets up all the graphics figures out how many fish are displayed,
 *    paints the fish and handles events.  The second is a seperate thread
 *    that calculates movement for the fish.  The calculations are done
 *    on a seperate thread in an attempt to not slow down the redrawing
 *    thread.  In general, since the fish won't be moving very fast there
 *    is no synchronization, if things are off they'll be right hopefully
 *    in 1/5 of a second...</p>
 *
 * @author <a href="mailto:jdw@jdwsoftware.com">Jim Woodgate</a>
 * @version $Revision: 1.8 $
 */

/*
 * This file is part of Woodys Fishing Hole.
 *
 * Woodys Fishing Hole is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * Woodys Fishing Hole is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Woodys Fishing Hole; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

public class GameComponent extends JComponent {
  private static final String COPYRIGHT = com.jdwsoftware.util.Copyright.COPYRIGHT;
  private static final String VERSION = "$Id: GameComponent.java,v 1.8 2003/12/10 05:29:47 jdw Exp $";

  private static final long DELAY   = 5l;
  private static final int GT_COLOR = 0;
  private static final int GT_NUM   = 1;
  private static final int GT_MATH  = 2;
  private static final int MAX_FISH = 7;
  private static final int DEFAULT_NUM_FISH = 3;

  private static final String[] GAME_TYPES = new String[] {"Colors",
                                                           "Numbers",
                                                           "Math"};

  private static final float CEILING = 50.0f;
  private boolean _paused = false;

  private Image _background = new ImageIcon(MainFrame.class.getResource("background.png")).getImage();
  private Image _hook       = new ImageIcon(MainFrame.class.getResource("hook.png")).getImage();
  private Image[] _fishImages = new Image[] {
      new ImageIcon(MainFrame.class.getResource("fish11.png")).getImage(),
      new ImageIcon(MainFrame.class.getResource("fish12.png")).getImage(),
      new ImageIcon(MainFrame.class.getResource("fish13.png")).getImage(),
      new ImageIcon(MainFrame.class.getResource("fish14.png")).getImage(),
      new ImageIcon(MainFrame.class.getResource("fish15.png")).getImage(),
      new ImageIcon(MainFrame.class.getResource("fish16.png")).getImage(),
      new ImageIcon(MainFrame.class.getResource("fish22.png")).getImage(),
      new ImageIcon(MainFrame.class.getResource("fish23.png")).getImage(),
      new ImageIcon(MainFrame.class.getResource("fish24.png")).getImage(),
      new ImageIcon(MainFrame.class.getResource("fish25.png")).getImage(),
      new ImageIcon(MainFrame.class.getResource("fish31.png")).getImage(),
      new ImageIcon(MainFrame.class.getResource("fish32.png")).getImage(),
      new ImageIcon(MainFrame.class.getResource("fish33.png")).getImage(),
      new ImageIcon(MainFrame.class.getResource("fish34.png")).getImage(),
      new ImageIcon(MainFrame.class.getResource("fish35.png")).getImage()
  };

  private Image[] _upFishImages;

  private Point mouseLoc = new Point();
  private Random r = new Random();
  private Fish[] _fishes;
  private Fish   _enteringFish;
  private Fish   _exittingFish;
  private JLabel _statusBar;

  private int _width = 640;
  private int _height = 480;
  private int _fishWidth;
  private int _fishHeight;
  private int _gameType = GT_COLOR;
  private int _numFish = DEFAULT_NUM_FISH;

  private GameType _currentGame = null;

  private CalculationThread _calcThread = null;

  /**
   * Default constructor
   */
  public GameComponent() {
    setOpaque(false);
    setSize(_width,_height);
    Cursor cursor;

    try {
      cursor = Toolkit.getDefaultToolkit().createCustomCursor(_hook,
          new Point(5, 26), "Hook");
    }
    catch (Exception ex) {
      cursor = Toolkit.getDefaultToolkit().createCustomCursor(_hook, new Point(1,1), "Hook");
    }
    setCursor(cursor);

    _fishWidth = _fishImages[0].getWidth(null);
    _fishHeight = _fishImages[0].getHeight(null);

    // set up the upfish images
    _upFishImages = new Image[_fishImages.length];

    for (int i=0; i<_fishImages.length; i++) {
      int[] buffer = new int[_fishWidth*_fishHeight];
      int[] rotate = new int[_fishWidth*_fishHeight];

      PixelGrabber grabber = new PixelGrabber(_fishImages[i], 0, 0, _fishWidth,
                                              _fishHeight, buffer, 0, _fishWidth);
      try {
        grabber.grabPixels();
        for(int x = 0; x < _fishWidth; x++) {
          for(int y = 0; y < _fishHeight; y++) {
            rotate[(x+1)*_fishHeight-y-1] = buffer[(y*_fishWidth)+x];
            }
          }
        _upFishImages[i] = createImage(new MemoryImageSource(_fishHeight, _fishWidth, rotate, 0, _fishHeight));
      }
      catch(InterruptedException e) { /* NOOP */ }


    }

    // Start off with 3 fishes
    _fishes = new Fish[MAX_FISH];
    for (int i=0; i<MAX_FISH; i++) {
      _fishes[i] = new Fish();
    }
    setNumFish(3);

    _calcThread = new CalculationThread();
    _calcThread.start();
  }

  /**
   * If the main component has a status bar, it should register that
   * status bar here so messages can be displayed.
   *
   * @param statusBar Label for displaying messages
   */
  public void setStatusBar(JLabel statusBar) {
    _statusBar = statusBar;
  }

  private void loadCurrentGame() {
    if (_exittingFish != null
        || _enteringFish != null) {
      return;
    }

    if (_currentGame == null) {
      switch (_gameType) {
        case GT_COLOR: _currentGame = new ColorGame(this);
                       break;
        case GT_NUM: _currentGame = new NumberGame(this);
                     break;
        case GT_MATH: _currentGame = new MathGame(this);
                      break;

        default: _currentGame = new ColorGame(this);
      }

    }
    else {
      _currentGame.newQuestion();
    }

    boolean foundAnswer = false;
    String[] answers = _currentGame.getValidAnswers();

    for (int i=0; i<_numFish; i++) {
      _fishes[i].setValue(answers[r.nextInt(answers.length)]);
      if (_fishes[i].getValue().equals(_currentGame.getAnswer())) {
        foundAnswer = true;
      }
    }
    if (!foundAnswer) {
      _fishes[r.nextInt(_numFish)].setValue(_currentGame.getAnswer());
    }
  }

  public synchronized void pauseGame(boolean pause) {
    _paused = pause;
  }

  protected void paintComponent(Graphics g) {
    if (_paused)
      return;

    if (_currentGame == null) {
      loadCurrentGame();
    }

    Image offscreen = this.createImage(_width, _height);

    Graphics offscreenG = offscreen.getGraphics();

    offscreenG.drawImage(_background,0, 0, null);

    // Draw what we're looking for
    if (_exittingFish == null
        && _enteringFish == null
        && _currentGame != null) {
      offscreenG.drawImage(_currentGame.getQuestion(),
                           _width/2-_currentGame.getQuestion().getWidth(null)/2,
                           5, null);
    }

    if (_exittingFish != null) {
      offscreenG.drawImage(_upFishImages[_exittingFish.getImageIndex()],
                           (int) _exittingFish.getLocation().getX(),
                           (int) _exittingFish.getLocation().getY(), null);

    }
    if (_enteringFish != null) {
      offscreenG.drawImage(_upFishImages[_enteringFish.getImageIndex()],
                           (int) _enteringFish.getLocation().getX()+_fishHeight,
                           (int) _enteringFish.getLocation().getY()+_fishWidth,
                           -_fishHeight,
                           -_fishWidth,
                           null);

    }
    for (int i=0; i<_numFish; i++) {
      if (_fishes[i] != null) {
        if (_fishes[i].getDirection().getDX() < 0.0f) {
          offscreenG.drawImage(_fishImages[_fishes[i].getImageIndex()],
                               (int) _fishes[i].getLocation().getX(),
                               (int) _fishes[i].getLocation().getY(), null);
        }
        else {
          offscreenG.drawImage(_fishImages[_fishes[i].getImageIndex()],
                               (int) _fishes[i].getLocation().getX() +
                               _fishWidth,
                               (int) _fishes[i].getLocation().getY(),
                               -_fishWidth, _fishHeight, null);
        }

        if (_exittingFish == null && _enteringFish == null) {

          // draw the value
          if ("UNSET".equals(_fishes[i].getValue())) {
            loadCurrentGame();
          }

          if (_currentGame == null) {
            loadCurrentGame();
          }

          offscreenG.drawImage(_currentGame.getAnswer(_fishes[i].getValue()),
                               (int) _fishes[i].getLocation().getX() +
                               (_fishWidth / 2) - _currentGame.getWidth() / 2,
                               (int) _fishes[i].getLocation().getY() +
                               (_fishHeight / 2) - _currentGame.getHeight() / 2,
                               _currentGame.getWidth(),
                               _currentGame.getHeight(), null);
        }
      }
    }

    offscreenG.setColor(Color.gray);
    offscreenG.drawLine(mouseLoc.x+8, mouseLoc.y-20, 320, -100);

    g.drawImage(offscreen,0,0, null);
    repaint();
  }

  /**
   * Get the preferred size of this component.
   *
   * @return preferred size
   */
  public Dimension getPreferredSize() {
      return new Dimension(_width,_height);
  }

  /**
   * Keep track of the current mouse location by being notified
   * of mouse movements here.
   *
   * @param p Where the mouse was last moved to
   */
  public void notifyMouseMoved(Point p) {
    mouseLoc.x = p.x;
    mouseLoc.y = p.y;

  }

  /**
   * Set the game type and then invalidate the current
   * game so a new game can be created
   *
   * @param gameType Type of game
   */
  public void setGameType(String gameType) {
    _gameType = -1;
    for (int i=0; i<GAME_TYPES.length; i++) {
      if (GAME_TYPES[i].equals(gameType)) {
        _gameType = i;
        break;
      }
    }

    if (_gameType == -1) {
      _gameType = GT_COLOR;
      if (_statusBar != null) {
        _statusBar.setText("Invalid Game Type " + gameType);
      }
    }
    if (_statusBar != null) {
      _statusBar.setText("Game type set to " + gameType);
      _currentGame = null;
    }
  }

  /**
   * Handle a mouse click.  If the click happens on a fish with the
   * correct answer, start the correct answer actions.
   *
   * @param p Point where the mouse was clicked
   */
  public void notifyMouseClicked(Point p) {
    // check for fish at the click
    for (int i=0; i<_numFish; i++) {
      if (_fishes[i] != null) {
        if (p.x >= _fishes[i].getLocation().getX()
            && p.x <= _fishes[i].getLocation().getX() + _fishWidth
            && p.y >= _fishes[i].getLocation().getY()
            && p.y <= _fishes[i].getLocation().getY() + _fishHeight) {
          if (_fishes[i].getValue().equals(_currentGame.getAnswer())) {
            if (_statusBar != null) {
              _statusBar.setText("Fish " + i + " having value " +
                                 _fishes[i].getValue() + " matches!");
            }
            _exittingFish = _fishes[i];
            if (_exittingFish.getDirection().getDX() > 0.0f) {
              _exittingFish.getLocation().setX(_exittingFish.getLocation().getX()+(_fishWidth-_fishHeight));
            }
            _fishes[i] = null;
          }
        }
      }
    }
  }

  /**
   * Get the current number of fish being displayed
   *
   * @return number of fish
   */
  public int getNumFish() {
    return _numFish;
  }

  /**
   * Set the number of fish to display.  Also resets any current
   * fish.
   *
   * @param numFish number of fish to use
   */
  public void setNumFish(int numFish) {
    if (numFish >= DEFAULT_NUM_FISH && numFish <= MAX_FISH) {
      _numFish = numFish;

      // set new values for all the fish
      for (int i=0; i<_numFish; i++) {
        if (_fishes[i] != null) {
          Location location = new Location(r.nextFloat() * (_width - _fishWidth),
                                           (CEILING +
                                            r.nextFloat() *
                                            (_height - _fishHeight - CEILING)));
          Direction direction = new Direction( (5.0f - r.nextFloat() * 10.0f),
                                              (1.5f - r.nextFloat() * 3.0f));
          int imageIndex = r.nextInt(_fishImages.length);

          _fishes[i].setValue("UNSET");
          _fishes[i].setDirection(direction);
          _fishes[i].setLocation(location);
          _fishes[i].setImageIndex(imageIndex);
        }
      }
    }
  }

  private class CalculationThread extends Thread {

    public void run() {
      while (true) {
        for (int i=0; i<_numFish; i++) {
          if (_fishes[i] != null) {
            if (mouseLoc.x < _fishes[i].getLocation().getX()
                || (mouseLoc.x > _fishes[i].getLocation().getX() + _fishWidth)
                || mouseLoc.y < _fishes[i].getLocation().getY()
                || (mouseLoc.y > _fishes[i].getLocation().getY() + _fishHeight)) {

              if (r.nextInt(1000) < 5) {
                _fishes[i].getDirection().setDX( -_fishes[i].getDirection().
                                                getDX());
              }

              if (r.nextInt(1000) < 5) {
                _fishes[i].getDirection().setDY( -_fishes[i].getDirection().
                                                getDY());
              }

              if (r.nextInt(2000) < 5) {
                Direction direction = new Direction((5.0f-r.nextFloat()*10.0f), (1.5f-r.nextFloat()*3.0f));
                _fishes[i].setDirection(direction);
              }

              _fishes[i].getLocation().setX(_fishes[i].getLocation().getX() +
                                            _fishes[i].getDirection().getDX());
              if (_fishes[i].getLocation().getX() < 0.0) {
                _fishes[i].getDirection().setDX( Math.abs(_fishes[i].getDirection().
                                                getDX()));
                _fishes[i].getLocation().setX(_fishes[i].getLocation().getX() +
                                              2 *
                                              _fishes[i].getDirection().getDX());
              }
              if (_fishes[i].getLocation().getX() > (_width - _fishWidth)) {
                _fishes[i].getDirection().setDX( -(Math.abs(_fishes[i].getDirection().
                                                getDX())));
                _fishes[i].getLocation().setX(_fishes[i].getLocation().getX() +
                                              2 *
                                              _fishes[i].getDirection().getDX());
              }

              _fishes[i].getLocation().setY(_fishes[i].getLocation().getY() +
                                            _fishes[i].getDirection().getDY());
              if (_fishes[i].getLocation().getY() < CEILING) {
                _fishes[i].getDirection().setDY( Math.abs(_fishes[i].getDirection().
                                                getDY()));
                _fishes[i].getLocation().setY(_fishes[i].getLocation().getY() +
                                              2 *
                                              _fishes[i].getDirection().getDY());
              }
              if (_fishes[i].getLocation().getY() > (_height - _fishHeight)) {
                _fishes[i].getDirection().setDY( -(Math.abs(_fishes[i].getDirection().
                                                getDY())));
                _fishes[i].getLocation().setY(_fishes[i].getLocation().getY() +
                                              2 *
                                              _fishes[i].getDirection().getDY());
              }
            }
          }
        }
        if (_exittingFish != null) {
          if (_exittingFish.getLocation().getY() > -200.0f) {
            _exittingFish.getLocation().setY(_exittingFish.getLocation().getY()-10.0f);
          }
          else {
            _enteringFish = _exittingFish;
            _exittingFish = null;
          }
        }
        if (_enteringFish != null) {
          if (_enteringFish.getLocation().getY() < 125.0f) {
            _enteringFish.getLocation().setY(_enteringFish.getLocation().getY()+10.0f);
          }
          else {
            for (int i=0; i<_numFish; i++) {
              if (_fishes[i] == null) {
                _fishes[i] = _enteringFish;
                _enteringFish = null;
                loadCurrentGame();
                break;
              }
            }
          }
        }

        // Take a break
        try {
          Thread.sleep(50l);
        }
        catch (InterruptedException e) { /* NOOP */ }
      }
    }
  }
}
